#
# flowrepo cms
#
# Requires djpcms
#
from django.template import loader
from django.contrib.contenttypes.models import ContentType
from django.forms.models import modelform_factory
from django import forms, http
from django.utils.encoding import force_unicode

from tagging.models import Tag
from tagging.forms import TagField

from flowrepo.models import FlowRelated, FlowItem, Report, Message, Category, Image
from flowrepo.forms import FlowItemSelector, FlowItemForm, ChangeCategory, ChangeImage, UloaderForm

from djpcms.views import appview, appsite
from djpcms.views.apps.tagging import tagurl
from djpcms.utils.html import form, formlet, htmlwrap, Paginator
from djpcms.utils.html import FormHelper, FormLayout, Fieldset, AutocompleteManyToManyInput
from djpcms.utils import form_kwargs
from djpcms.plugins import DJPplugin


def AutocompleteTagField(required = False):
    wg = AutocompleteManyToManyInput(Tag, ['name'], separator = ' ', inline = True)
    return TagField(required = required, widget = wg)


class ContentView(appview.SearchView):
    
    def __init__(self, *args, **kwargs):
        super(ContentView,self).__init__(*args,**kwargs)
        
    def title(self, page, content = None, **urlargs):
        if content:
            cmodels  = self.appmodel.content_models
            contents = content.split('+')
            titles   = []
            for content in contents:
                model = cmodels.get(content,None)
                if model:
                    titles.append(force_unicode(model._meta.verbose_name_plural))
            return '+'.join(titles)
        else:
            return super(ContentView,self).title(page,**urlargs)
        
    def appquery(self, request, content = None, **kwargs):
        qs = self.basequery(request, **kwargs)
        if content:
            cmodels  = self.appmodel.content_models
            contents = content.split('+')
            ids      = []
            for content in contents:
                model = cmodels.get(content,None)
                if model:
                    ctype = ContentType.objects.get_for_model(model)
                    ids.append(ctype.id)
            if ids:
                return qs.filter(content_type__pk__in = ids)
            else:
                return self.model.objects.none()
        raise self.model.objects.none()
    

class NiceUloaderForm(UloaderForm):
    tags         = AutocompleteTagField()
    
    helper = FormHelper()
    helper.ajax = False
    
    # add the layout object
    helper.layout = FormLayout(
                             Fieldset('visibility', 'tags', 'name',
                                      css_class = Fieldset.inlineLabels)
                             )

       
class FlowItemApplication(tagurl.ArchiveTaggedApplication):
    '''
    Base application class for flowitem models
    '''
    date_code        = 'timestamp'
    form_withrequest = True
    form             = FlowItemForm
    split_days       = True
    inherit          = True
    insitemap        = True
    content_names    = {}   # dictionary for mapping a model into a name used for url
    
    upload_file      = appview.AddView(regex = 'upload', form = NiceUloaderForm) 
    
    def __init__(self, *args, **kwargs):
        super(FlowItemApplication,self).__init__(*args, **kwargs)
        content_models = {}
        for cn,model in FlowItem.objects.models_by_name.items():
            cname = self.content_names.get(model,None)
            if not cname:
                cname = cn
                self.content_names[model] = cname
            content_models[cname] = model
        self.content_models = content_models
        
    def objptr(self, object):
        try:
            return object.object
        except:
            return object
    
    def flowitem(self, object):
        try:
            return object.flowitem()
        except:
            return None

    def modelsearch(self):
        return FlowItem
            
    def basequery(self, request):
        return FlowItem.objects.public(user = request.user, model = self.model)
    
    def get_search_fields(self):
        if self.search_fields:
            return self.search_fields
        else:
            from django.contrib.admin import site
            admin = site._registry.get(FlowItem,None)
            if admin:
                return admin.search_fields

    def has_permission(self, request = None, obj=None):
        '''
        evaluate only if obj is an instance of flowitem
        '''
        obj = self.flowitem(obj)
        if not obj:
            return True
        if not request:
            return obj.can_user_view()
        else:
            return obj.can_user_view(request.user)
    
    def has_edit_permission(self, request = None, obj=None):
        obj = self.flowitem(obj)
        if super(FlowItemApplication,self).has_edit_permission(request, obj):
            if obj:
                return request.user in obj.authors.all()
            else:
                return True
        else:
            return False
        
    def get_object_view_template(self, obj, wrapper):
        item = self.flowitem(obj)
        obj  = item.object
        opts = obj._meta
        template_name = '%s.html' % opts.module_name
        return ['%s/%s' % (opts.app_label,template_name),
                'djpcms/components/object.html']
        
    def get_item_template(self, obj, wrapper):
        model = self.objptr(obj)._meta.module_name
        return ['components/%s_list_item.html' % model,
                'flowrepo/%s_list_item.html' % model,
                'flowrepo/flowitem_list_item.html']
    
    def title_object(self, obj):
        object = self.objptr(obj)
        name = getattr(object,'name',None)
        if not name:
            return str(obj)
        else:
            return name
        
    def object_content(self, djp, obj):
        '''
        Utility function for getting more content out of an instance of a model
        '''
        c = super(FlowItemApplication,self).object_content(djp,obj)
        c['authors']  = obj.niceauthors()
        c['external'] = True
        return c
    
    def objectbits(self, obj):
        '''
        Get arguments from model instance used to construct url
        By default it is the object id
        @param obj: instance of self.model
        @return: dictionary of url bits 
        '''
        obj = obj.object
        rep = {'content':self.content_names[obj.__class__]}
        if hasattr(obj,'slug'):
            rep['id'] = obj.slug
        else:
            rep['id'] = obj.id
        return rep
    
    def get_object(self, *args, **kwargs):
        '''
        Retrive an instance of self.model for arguments.
        By default arguments is the object id,
        Reimplement for custom arguments
        '''
        try:
            cn = kwargs.get('content')
            model = self.content_models.get(cn,None)
            if model:
                id   = kwargs.get('id',None)
                try:
                    obj = model.objects.get(slug = id)
                except:
                    try:
                        obj = model.objects.get(id = int(id))
                    except:
                        return None
            return FlowItem.objects.get_from_instance(obj)
        except:
            return None

    def viewurl(self, request = None, obj = None):
        if self.model is FlowItem:
            model = obj.object.__class__
            appmodel = appsite.site.for_model(model)
            if appmodel:
                return appmodel.viewurl(request,obj)
        return super(FlowItemApplication,self).viewurl(request,obj)
        
    def editurl(self, request = None, obj = None):
        if self.model is FlowItem:
            model = obj.object.__class__
            appmodel = appsite.site.for_model(model)
            if appmodel:
                return appmodel.editurl(request,obj)
        return super(FlowItemApplication,self).editurl(request,obj)



class ReportApplication(FlowItemApplication):
    inherit  = True
    form_layout = 'flowrepo/report_form.html'
    
    add      = appview.AddView(regex = 'write')
    edit     = appview.EditView(regex = '(?P<id>[-\.\w/]+)', parent = 'add', splitregex = False)
    view     = appview.ViewView(regex = '(?P<id>[-\.\w/]+)', splitregex = False)
            
    def objectbits(self, object):
        '''
        Get arguments from model instance used to construct url
        By default it is the object id
        @param obj: instance of self.model
        @return: dictionary of url bits 
        '''
        obj = self.objptr(object)
        if hasattr(obj,'slug'):
            parent = getattr(obj,'parent',None)
            if parent:
                bits = '%s/%s' % (self.objectbits(parent)['id'],obj.slug)
            else:
                bits = obj.slug
            return {'id': bits}
        else:
            return {'id': obj.id}
    
    def get_object(self, *args, **kwargs):
        '''
        Retrive an instance of self.model for arguments.
        By default arguments is the object id,
        Reimplement for custom arguments
        '''
        try:
            id   = kwargs.get('id',None)
            try:
                slug = id.split('/')[-1]
                obj = self.model.objects.get(slug = slug)
            except:
                try:
                    obj = self.model.objects.get(id = int(id))
                except:
                    return None
            return FlowItem.objects.get_from_instance(obj)
        except:
            return None
        
    def parentresponse(self, djp, app):
        obj = self.objptr(djp.instance)
        if hasattr(obj,'slug'):
            parent = getattr(obj,'parent',None)
            if parent:
                return app(djp.request, instance = FlowItem.objects.get_from_instance(parent))
        return super(ReportApplication, self).parentresponse(djp, app)


#
#
#____________________________________________ DJPCMS PLUGINS

class LinkedItemForm(forms.Form):
    content_types = forms.ModelMultipleChoiceField(ContentType.objects.all(),
                                                   required = False,
                                                   label = 'model')
    display = forms.ChoiceField(choices = (('list','list'),
                                           ('detail','detail')), initial = 'list')
    inverse = forms.BooleanField(initial = False, required = False)


class FlowItemSelection(DJPplugin):
    name = 'flowitem-selection'
    description = 'Items Selection'
    form = FlowItemSelector
    
    def render(self, djp, wrapper, prefix,
               visibility = None, item_per_page = 10,
               content_type = None, tags = '',
               **kwargs):
        request = djp.request
        qs      = FlowItem.objects.selection(request.user,
                                             types = content_type,
                                             visibility = visibility,
                                             tags = tags)
        if not qs:
            return None
        pa = Paginator(djp.request, qs, item_per_page)
        return loader.render_to_string(['flowitem_list.html',
                                        'djpcms/components/object_list.html'],
                                        {'items': self.paginator(djp,pa)})
        
    def paginator(self, djp, pa):
        from djpcms.views import appsite
        appmodel  = appsite.site.for_model(FlowItem)
        qs        = pa.qs
        for obj in qs:
            object    = obj.object
            model     = object.__class__
            objmodel  = appsite.site.for_model(model) or appmodel
            if objmodel:
                content = objmodel.object_content(djp, obj)
                tname   = '%s_list_item.html' % model.__name__.lower()
                yield loader.render_to_string(['components/%s' % tname,
                                               'flowrepo/%s' % tname,
                                               'flowrepo/flowitem_list_item.html'],
                                               content)


class ImagePlugin(DJPplugin):
    '''
    Plugin for displaying a single image on a content block
    '''
    name = 'image'
    description = 'Display Image'
    form = ChangeImage
    
    def render(self, djp, wrapper, prefix, image = None, class_name = None, **kwargs):
        try:
            img = Image.objects.get(id = int(image))
            if class_name:
                class_name = '%s image-plugin'
            else:
                class_name = 'image-plugin'
            cn = ' class="%s"' % class_name
            return '<img src="%s" title="%s" alt="%s"%s/>' % (img.url,img.name,img.name,cn)
        except:
            return u''
    

class CategoryLinks(DJPplugin):
    name = 'category-links'
    description = "Display links for a category"
    form          = ChangeCategory
    
    def render(self, djp, wrapper, prefix, category_name = None, **kwargs):
        if category_name:
            qs = Category.objects.filter(type__id = int(category_name))
        else:
            return u''
        if not qs:
            return u''
        return loader.render_to_string(['category_list.html',
                                        'components/category_list.html',
                                        'flowrepo/category_list.html'],
                                        {'items': self.paginator(djp,qs)})
    
    def paginator(self, djp, qs):
        from djpcms.views import appsite
        appmodel = appsite.site.for_model(Category)
        for obj in qs:
            if appmodel:
                content  = appmodel.object_content(djp, obj)
                yield loader.render_to_string(['category_list_item.html',
                                               'components/category_list_item.html',
                                               'flowrepo/category_list_item.html'],
                                               content)
        
    def edit_form(self, djp, category_name = None, **kwargs):
        if category_name:
            return EditContentForm(**form_kwargs(request = djp.request,
                                                 initial = {'category_name': category_name},
                                                 withrequest = True))


class LinkedItems(DJPplugin):
    '''
    Very very useful plugin for displaying a list of linked item
    to the flowitem object
    '''
    name = 'flowrepo-linked'
    description = "Related Items"
    form = LinkedItemForm
    
    def render(self, djp, wrapper, prefix, content_types = None, display = 'list', inverse = False, **kwargs):
        from djpcms.views import appsite
        instance = djp.instance
        if not isinstance(instance,FlowItem):
            return
        qs = instance.items.all()
        lqs = list(qs)
        if content_types:
            qs = qs.filter(content_type__id__in = content_types)
        if not qs:
            return ''
        request = djp.request
        return loader.render_to_string(['report_draft_list.html',
                                        'flowrepo/related_list.html'],
                                        {'items': self.paginator(djp,wrapper,qs,display)})
    
    def paginator(self, djp, wrapper, qs, display):
        for obj in qs:
            object  = obj.object
            opts    = object._meta
            appmodel = appsite.site.for_model(object.__class__)
            
            if display == 'detail':
                if appmodel:
                    yield appmodel.render_object(djp, wrapper)
                else:
                    yield loader.render_to_string(['%s/%s.html' % (opts.app_label,opts.module_name),
                                                   'djpcms/components/object.html'],
                                                   {'item':object})
            
            url     = None
            r       = ''
            try:
                url     = object.url
            except:
                appmodel = appsite.site.for_model(object.__class__)
                if appmodel:
                    url     = appmodel.viewurl(djp.request,object)
            if url:
                try:
                    name = object.name
                    if not name:
                        name = str(object)
                except:
                    name = str(object)
                link    = htmlwrap('a', name)
                link._attrs['title'] = name
                link._attrs['href'] = url
                if not url.startswith('/'):
                    link._attrs['target'] = "_blank"
                r = link.render()
            
            yield r


class AddLinkedItems(DJPplugin):
    name = 'flowrepo-add-linked'
    description = "Add Linked item urls"
    
    def render(self, djp, wrapper, prefix, **kwargs):
        from djpcms.views import appsite
        instance = djp.instance
        if not isinstance(instance,FlowItem):
            return
        qs = FlowRelated.objects.filter(item = instance)
        if not qs:
            return
        request = djp.request
        return loader.render_to_string(['report_draft_list.html',
                                        'flowrepo/report_draft_list.html'],
                                        {'items': self.paginator(djp,qs)})

