from django import forms
from django.contrib.auth.models import User
from django.utils.encoding import force_unicode
import django.utils.simplejson as json

from tagging.forms import TagField

from flowrepo import models


class FlowItemSelector(forms.ModelForm):
    '''
    Form for selecting items to display
    '''
    content_type = forms.ModelMultipleChoiceField(queryset = models.FlowItem.objects.allmodels(),
                                                  label = 'types',
                                                  required = False)
    item_per_page = forms.IntegerField(initial = 10)
    
    class Meta:
        model = models.FlowItem
        fields = ('content_type', 'visibility', 'tags')
        
        
    def save(self, commit = True):
        pass


class FlowFormBase(forms.ModelForm):
    tags           = TagField(required = False)
    visibility     = forms.ChoiceField(choices=models.visibility_choices, initial=3)
    
    def clean_visibility(self):
        return int(self.cleaned_data['visibility'])
    
    def __init__(self, *args, **kwargs):
        user = kwargs.pop('user',None)
        if not user:
            request = kwargs.pop('request',None)
            if request:
                user = request.user
        self.user = user
        item = kwargs.get('instance',None)
        if item:
            initial = kwargs.pop('initial',None) or {}
            if isinstance(item,models.FlowItem):
                instance = item.object
            else:
                try:
                    instance = item
                    item = instance.flowitem()
                except:
                    instance = None
                    item = None
            kwargs['instance']        = instance
            if item:
                initial['tags']           = item.tags
                initial['timestamp']      = item.timestamp
                initial['visibility']     = force_unicode(item.visibility)
                initial['allow_comments'] = item.allow_comments
                kwargs['initial']         = initial
                for name,field in self.declared_fields.items():
                    try:
                        model = field.queryset.model
                    except:
                        continue
                    if model == User:
                        initial[name] = [user.id for user in item.authors.all()]
                    relems = item.items.filter_for_model(model)
                    if relems:
                        initial[name] = [rel.object_id for rel in relems]
            
        super(FlowFormBase, self).__init__(*args, **kwargs)
    
    def commitsave(self, instance, user, registered):
        '''
        save model and if registered with FlowItem, save the pointer
        '''
        instance.save()
        if registered:
            flowitems  = models.FlowItem.objects
            flowitems.create_or_update(instance, user = user, **self.cleaned_data)
            # Add related item if available in the form
            item = instance.flowitem()
            if item:
                for name,field in self.declared_fields.items():
                    if name not in self.changed_data:
                        continue
                    try:
                        model = field.queryset.model
                    except:
                        continue
                    data = self.cleaned_data.get(name,None)
                    if model == User:
                        item.authors = data
                        item.authors.add(user)
                        item.save()
                    else:
                        item.items.filter_for_model(model).delete()
                        if data:
                            item.update_related(data)
        return instance
            
    def save(self, commit = True):
        model      = self.get_model()
        flowitems  = models.FlowItem.objects
        registered = flowitems.unregisterModel(model)
        instance   = super(FlowFormBase,self).save(commit = commit)
        if registered:
            if commit:
                self.instance = self.commitsave(instance, self.user, registered)
            flowitems.registerModel(model)
        return instance
    
    def save_model(self, instance, user):
        model      = self.get_model()
        flowitems  = models.FlowItem.objects
        registered = flowitems.unregisterModel(model)
        self.instance = self.commitsave(instance, user, registered)
        if registered:
            flowitems.registerModel(model)
        return self.instance
    
    def get_model(self):
        return self._meta.model
    

class FlowItemForm(FlowFormBase):
    '''
    General form for a flowrepo model
    '''
    timestamp      = forms.DateTimeField(required = False)
    allow_comments = forms.BooleanField(required = False, initial = False)
    

def get_upload_model(file):
    '''Upload model'''
    try:
        from PIL import Image
    except ImportError:
        import Image

    model = models.Attachment
    try:
        if file.content_type.split('/')[0] == 'image':
            model = models.Image
        #trial_image = Image.open(file)
        #trial_image.load()
        #trial_image.verify()
        #model = models.Image
    except:
        model = models.Attachment
    
    return model
        
        
class UloaderForm(FlowFormBase):
    file = forms.FileField(help_text = 'File to upload')
    
    def __init__(self, *args, **kwargs):
        files = kwargs.get('files',None)
        if files:
            file = files.get('file',None)
            if file:
                self.specify_model(file)
        super(UloaderForm,self).__init__(*args, **kwargs)
        
    class Meta:
        exclude = ['authors','groups','timestamp','allow_comments','url']
            
    def clean_file(self):
        elem = self.cleaned_data.get('file')
        if not file:
            url = self.cleaned_data.get('elem_url')
            if not url:
                raise forms.ValidationError('Remote url or file must be available')
        return elem
    
    def specify_model(self, file):
        model = get_upload_model(file)
        self._meta.model = model
        return model

    
class LinkedItemForm(forms.Form):
    pass



def get_flow_form(self, request, instance = None):
    '''
    This function get the form for editing/adding object pointed by item
    '''
    if instance == None:
        return None
    model = instance.content_type.model_class()
    return modelform_factory(model,FlowItemForm)
    if model == Report:
        return modelform_factory(model,FlowItemForm)
    else:
        return modelform_factory(model,FlowItemForm)
    




class WebAccountForm(forms.ModelForm):
    user = forms.CharField(widget=forms.HiddenInput, required = False)
    '''
    A form for web account.
    User must be injected when creating a new class instance using modelform_factory
    '''
    def __init__(self, user, **kwargs):
        self._user = user
        super(WebAccountForm,self).__init__(**kwargs)
        
    class Meta:
        model   = models.WebAccount
        exclude = ['e_data']
    
    def clean_user(self):
        return self._user
    
    def clean_name(self):
        user = self._user
        name = self.cleaned_data['name']
        if name:
            acc = self._meta.model.objects.filter(user = user, name = name)
            if acc:
                acc = acc[0]
                if acc.id != self.instance.id:
                    raise forms.ValidationError('Account %s already available' % name)
            return name
        else:
            raise forms.ValidationError('Account name is required')
    
    @classmethod
    def make(cls, user, data = None, instance = None, **kwargs):
        f1 = cls(user, data = data, instance = instance, **kwargs)
        f2 = DataForm(data = data, instance = instance)
        return (f1,f2)
    

class DataForm(forms.Form):
    '''
    Form to use in conjunction with WebAccountForm.
    This form contain the user login data which will be stored in an encrypted format.
    '''
    username = forms.CharField(required = False, max_length = 200)
    password = forms.CharField(required = False, max_length = 200)
    email    = forms.EmailField(required = False)
    pin      = forms.CharField(required = False, max_length = 200)
    extended = forms.CharField(required = False)
    
    def __init__(self, instance = None, **kwargs):
        initial = None
        if instance:
            data = instance.data
            if data:
                initial = json.loads(data)
        super(DataForm,self).__init__(initial = initial, **kwargs)
        
    def tojson(self):
        if self.is_valid():
            return json.dumps(self.cleaned_data)
        else:
            return u''
         


class ChangeCategory(forms.Form):
    category_name = forms.ModelChoiceField(queryset = models.CategoryType.objects.all(),
                                           empty_label=None)
    
class ChangeImage(forms.Form):
    image   = forms.ModelChoiceField(queryset = models.Image.objects.all(), empty_label=None)
    class_name = forms.CharField(max_length = 100, required = False)
    
def add_related_upload(file, instance, name = None, description = ''):
    item = instance.flowitem()
    if not item or not file:
        return
    model = get_upload_model(file)    
    name = name or file.name
    obj = model(elem = file, name = name, description = description)
    obj.save()
    return models.FlowRelated.objects.create_for_object(item,obj)
    
    
    
